<?php
/**
 * Copyright (C) 2017-2018 thirty bees
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Academic Free License (AFL 3.0)
 * that is bundled with this package in the file LICENSE.md
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/afl-3.0.php
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@thirtybees.com so we can send you a copy immediately.
 *
 * @author    thirty bees <contact@thirtybees.com>
 * @copyright 2017-2018 thirty bees
 * @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
 */

use VATNumberModule\VatStatus;

if (!defined('_TB_VERSION_')) {
    exit;
}

require_once __DIR__.'/vendor/autoload.php';

/**
 * Class VATNumber
 */
class VATNumber extends TaxManagerModule
{
    const COUNTRY = 'VATNUMBER_COUNTRY';
    const CHECK_ENABLED = 'VATNUMBER_CHECK_ENABLED';
    const MANAGEMENT = 'VATNUMBER_MANAGEMENT';
    const GROUPS_CONFIG = 'VATNUMBER_GROUPS_CONF';
    const CONNECTION_ATTEMPTS = 3;

    /** @var array $hooks */
    public $hooks = [
        'actionAdminCustomersListingFieldsModifier',
    ];

    /**
     * @var array $vatPrefixes
     */
    public static $vatPrefixes = [
        'AT' => [
            'code' => 'AT',
            'name' => 'Austria',
        ],
        'BE' => [
            'code' => 'BE',
            'name' => 'Belgium',
        ],
        'DK' => [
            'code' => 'DK',
            'name' => 'Denmark',
        ],
        'FI' => [
            'code' => 'FI',
            'name' => 'Finland',
        ],
        'FR' => [
            'code' => 'FR',
            'name' => 'France',
        ],
        'FX' => [
            'code' => 'FR',
            'name' => 'France métropolitaine',
        ],
        'DE' => [
            'code' => 'DE',
            'name' => 'Germany',
        ],
        'GR' => [
            'code' => 'EL',
            'name' => 'Greece',
        ],
        'IE' => [
            'code' => 'IE',
            'name' => 'Ireland',
        ],
        'IT' => [
            'code' => 'IT',
            'name' => 'Italy',
        ],
        'LU' => [
            'code' => 'LU',
            'name' => 'Luxembourg',
        ],
        'NL' => [
            'code' => 'NL',
            'name' => 'Netherlands',
        ],
        'PT' => [
            'code' => 'PT',
            'name' => 'Portugal',
        ],
        'ES' => [
            'code' => 'ES',
            'name' => 'Spain',
        ],
        'SE' => [
            'code' => 'SE',
            'name' => 'Sweden',
        ],
        'GB' => [
            'code' => 'GB',
            'name' => 'United Kingdom',
        ],
        'CY' => [
            'code' => 'CY',
            'name' => 'Cyprus',
        ],
        'EE' => [
            'code' => 'EE',
            'name' => 'Estonia',
        ],
        'HU' => [
            'code' => 'HU',
            'name' => 'Hungary',
        ],
        'LV' => [
            'code' => 'LV',
            'name' => 'Latvia',
        ],
        'LT' => [
            'code' => 'LT',
            'name' => 'Lithuania',
        ],
        'MT' => [
            'code' => 'MT',
            'name' => 'Malta',
        ],
        'PL' => [
            'code' => 'PL',
            'name' => 'Poland',
        ],
        'SK' => [
            'code' => 'SK',
            'name' => 'Slovakia',
        ],
        'CZ' => [
            'code' => 'CZ',
            'name' => 'Czechia',
        ],
        'SI' => [
            'code' => 'SI',
            'name' => 'Slovenia',
        ],
        'RO' => [
            'code' => 'RO',
            'name' => 'Romania',
        ],
        'BG' => [
            'code' => 'BG',
            'name' => 'Bulgaria',
        ],
        'HR' => [
            'code' => 'HR',
            'name' => 'Croatia',
        ],
    ];

    /**
     * VATNumber constructor.
     *
     * @throws PrestaShopException
     */
    public function __construct()
    {
        $this->name = 'vatnumber';
        $this->tab = 'billing_invoicing';
        $this->version = '2.0.4';
        $this->author = 'thirty bees';
        $this->need_instance = 0;

        $this->tax_manager_class = 'VATNumberTaxManager';

        $this->bootstrap = true;
        parent::__construct();

        if (!Configuration::get(static::COUNTRY) && Configuration::get(static::CHECK_ENABLED)) {
            $this->warning = $this->l('No default country set.');
        }

        $this->displayName = $this->l('European VAT number');
        $this->description = $this->l('Enables you to enter the intra-community VAT number when creating the address. You must fill in the company field to allow entering the VAT number.');
    }

    /**
     * @return bool
     *
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function install()
    {
        if (!parent::install()) {
            return false;
        }

        $this->installDb();
        $this->installTab();

        return true;
    }

    /**
     * @return bool
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public function uninstall()
    {
        $this->uninstallDB();
        $this->uninstallTab();

        return parent::uninstall();
    }

    /**
     * @throws PrestaShopException
     */
    protected function installDb()
    {
        VatStatus::createDatabase();
    }

    /**
     * Install the tab for this module
     *
     * @return bool Whether the tab was successfully installed
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    protected function installTab()
    {
        if (Tab::getIdFromClassName('AdminVatNumbers')) {
            return true;
        }
        $mainTab = (int) Db::getInstance(_PS_USE_SQL_SLAVE_)->getValue(
            (new DbQuery())
                ->select('t.`id_tab`')
                ->from('tab', 't')
                ->where('t.`class_name` = \'AdminCustomers\'')
        );
        if ($mainTab < 1) {
            $mainTab = 0;
        }

        $postnlCsvTab = new Tab();
        $postnlCsvTab->active = 1;
        $postnlCsvTab->class_name = 'AdminVatNumbers';
        $postnlCsvTab->name = [];
        foreach (Language::getLanguages(true) as $lang) {
            if (strtoupper($lang['iso_code']) === 'NL') {
                $name = 'Btw-nummers';
            } else {
                $name = 'VAT Numbers';
            }
            $postnlCsvTab->name[$lang['id_lang']] = $name;
        }
        $postnlCsvTab->id_parent = (int) $mainTab;
        $postnlCsvTab->module = $this->name;
        $postnlCsvTab->add();

        return true;
    }

    /**
     * @throws PrestaShopException
     */
    protected function uninstallDB()
    {
        VatStatus::dropDatabase();
    }

    /**
     * Uninstall the tab
     *
     * @return void
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    protected function uninstallTab()
    {
        $tabs = Tab::getCollectionFromModule($this->name)->getResults();

        /** @var Tab $tab */
        foreach ($tabs as $tab) {
            $tab->delete();
        }
    }

    /**
     * @param array $args
     *
     * @return bool
     *
     * @since   1.0.0
     * @version 1.0.0 Initial version
     */
    public function hookTaxManager($args)
    {
        $class = $this->tax_manager_class;
        if (call_user_func([$class, 'isAvailableForThisAddress'], $args['address'])) {
            return new $class();
        }

        return false;
    }

    /**
     * @return array
     */
    public function getPrefixIntracomVAT()
    {
        static $prefixes = null;
        if (!$prefixes) {
            $prefixes = array_combine(array_keys(static::$vatPrefixes), array_column(static::$vatPrefixes, 'code'));
        }

        return $prefixes;
    }

    /**
     * @param int $idCountry
     *
     * @return int
     * @throws PrestaShopException
     * @throws Adapter_Exception
     */
    public static function isApplicable($idCountry)
    {
        /** @var VATNumber $module */
        $module = Module::getInstanceByName('vatnumber');

        return (((int) $idCountry
            && $idCountry != Configuration::get(static::COUNTRY)
            && array_key_exists(Country::getIsoById($idCountry), $module->getPrefixIntracomVAT())) ? 1 : 0);
    }

    /**
     * @param string $vatNumber
     *
     * @return array
     * @throws Adapter_Exception
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public static function webServiceCheck($vatNumber)
    {
        if (empty($vatNumber)) {
            return [];
        }

        /** @var VATNumber $module */
        $module = Module::getInstanceByName('vatnumber');
        $vatNumber = static::filterVatNumber($vatNumber);
        $prefix = strtoupper(substr($vatNumber, 0, 2));
        if (!in_array($prefix, array_keys($module->getPrefixIntracomVAT()))) {
            return [Tools::displayError('Invalid VAT number')];
        }

        $results = static::viesCheck($vatNumber);
        if (!empty($results[$vatNumber]['valid'])) {
            if (debug_backtrace()[1]['function'] === 'validateController' && is_a(debug_backtrace()[1]['object'], 'AddressCore')) {
                $address = debug_backtrace()[1]['object'];
                if ($address->id_country == Configuration::get(static::COUNTRY)) {
                    return [Tools::displayError('VAT numbers have been disabled for this country')];
                }

                $status = VatStatus::getByCustomerId(Context::getContext()->customer->id);
                $status->id_customer = Context::getContext()->customer->id;
                $status->valid = true;
                $status->manual = false;
                $status->vat_number = $results[$vatNumber]['vatNumber'];
                $status->vat_company = $results[$vatNumber]['traderName'];
                $status->vat_address = $results[$vatNumber]['traderAddress'];
                if ($status->save()) {
                    $status->assignGroup();
                }
            }

            return [];
        }

        return [Tools::displayError('VAT number validation service unavailable')];
    }

    /**
     * @param string|array|array[] $vatNumbers A single VAT number, an array with company data or multiple companies
     *                                        VAT Numbers should contain the Country prefix (DK, NL, etc.)
     *
     * @return array
     * @throws Adapter_Exception
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     */
    public static function viesCheck($vatNumbers)
    {
        /** @var VATNumber $module */
        $module = Module::getInstanceByName('vatnumber');
        if (!is_array($vatNumbers)) {
            // This is just a single number
            $vatNumbers = [
                [
                    'vatNumber' => static::filterVatNumber($vatNumbers),
                ],
            ];
        } elseif (is_array($vatNumbers) && count($vatNumbers) && !is_array($vatNumbers[0])) {
            // This is a single company array
            $vatNumbers = [$vatNumbers];
        } elseif (is_array($vatNumbers) && count($vatNumbers) && !isset($vatNumbers['vatNumber'])) {
            // This is an array of VAT numbers
            $vatNumbers = array_map(function ($item) {
                return ['vatNumber' => $item];
            }, $vatNumbers);
        }

        $guzzle = new \GuzzleHttp\Client(['timeout' => 60]);
        $promises = [];
        foreach ($vatNumbers as $company) {
            $prefix = strtoupper(substr($company['vatNumber'], 0, 2));
            $number = strtoupper(substr($company['vatNumber'], 2));
            if (array_search($prefix, array_keys($module->getPrefixIntracomVAT())) === false) {
                return [
                    'valid'   => false,
                    'success' => false,
                    'message' => Tools::displayError('Invalid VAT number'),
                ];
            }
            $xml = '<?xml version="1.0" encoding="utf-8" ?>';
            $xml .= '<s11:Envelope xmlns:s11="http://schemas.xmlsoap.org/soap/envelope/"><s11:Body>';
            $xml .= '<tns1:checkVatApprox xmlns:tns1="urn:ec.europa.eu:taxud:vies:services:checkVat:types">';
            $xml .= "<tns1:countryCode>{$prefix}</tns1:countryCode><tns1:vatNumber>{$number}</tns1:vatNumber>";
            if (!empty($company['name'])) {
                $xml .= "<tns1:traderName>{$company['name']}</tns1:traderName>";
            }
            if (!empty($company['street'])) {
                $xml .= "<tns1:traderStreet>{$company['street']}</tns1:traderStreet>";
            }
            if (!empty($company['postcode'])) {
                $xml .= "<tns1:traderPostcode>{$company['postcode']}</tns1:traderPostcode>";
            }
            if (!empty($company['city'])) {
                $xml .= "<tns1:traderCity>{$company['city']}</tns1:traderCity>";
            }
            $xml .= '</tns1:checkVatApprox></s11:Body></s11:Envelope>';
            $promises[$company['vatNumber']] = $guzzle->postAsync(
                'http://ec.europa.eu/taxation_customs/vies/services/checkVatService',
                [
                    'body' => $xml,
                    'headers' => [
                        'Content-Type' => 'application/xml',
                    ],
                ]
            );
        }
        $results = \GuzzleHttp\Promise\settle($promises)->wait();
        foreach ($results as &$result) {
            if ($result instanceof Exception) {
                $result = [
                    'valid'   => false,
                    'succes'  => false,
                    'message' => $result->getMessage(),
                ];

                continue;
            }

            /** @var \GuzzleHttp\Psr7\Response $response */
            $response = $result['value'];
            $body = (string) $response->getBody();
            $xml = new SimpleXMLElement($body);
            $result = (array) $xml->xpath('/soap:Envelope/soap:Body')[0]->checkVatApproxResponse;
            unset($result['requestIdentifier']);
            $result['valid'] = ($result['valid'] === 'true');
        }

        return $results;
    }

    /**
     * Filter VAT Number
     *
     * @param string $vatNumber
     *
     * @return mixed
     */
    public static function filterVatNumber($vatNumber)
    {
        return str_replace(' ', '', $vatNumber);
    }

    /**
     * @return string
     * @throws Exception
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function getContent()
    {
        \VATNumberModule\VatStatus::createDatabase();

        if (Tools::isSubmit('submitVatNumber')) {
            if (Configuration::updateValue(static::COUNTRY, (int) Tools::getValue(static::COUNTRY))) {
                $this->context->controller->confirmations[] = $this->l('Your country has been updated.');
            }

//            if (Configuration::updateValue(static::CHECK_ENABLED, (int) Tools::getValue(static::CHECK_ENABLED))) {
//                $this->context->controller->confirmations[] = (Tools::getValue(static::CHECK_ENABLED)
//                    ? $this->displayConfirmation($this->l('The check of the VAT number with the WebService is now enabled.'))
//                    : $this->displayConfirmation($this->l('The check of the VAT number with the WebService is now disabled.')));
//            }
        }

        return $this->renderForm().
            $this->renderVatGroupsList();
    }

    /**
     * @return string
     *
     * @throws Exception
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public function renderForm()
    {
        $countries = Country::getCountries($this->context->language->id);

        $countries_fmt = [
            0 => [
                'id'   => 0,
                'name' => $this->l('-- Choose a country --'),
            ],
        ];

        foreach ($countries as $country)
            $countries_fmt[] = [
                'id'   => $country['id_country'],
                'name' => $country['name'],
            ];

        $fieldsForm = [
            'form' => [
                'legend' => [
                    'title' => $this->l('Settings'),
                    'icon'  => 'icon-cogs',
                ],
                'input'  => [
                    [
                        'type'          => 'select',
                        'label'         => $this->l('Customers\' country'),
                        'desc'          => $this->l('Filter customers\' country.'),
                        'name'          => static::COUNTRY,
                        'required'      => false,
                        'default_value' => (int) $this->context->country->id,
                        'options'       => [
                            'query' => $countries_fmt,
                            'id'    => 'id',
                            'name'  => 'name',
                        ],
                    ],
//                    [
//                        'type'    => 'switch',
//                        'label'   => $this->l('Enable checking of the VAT number with the web service'),
//                        'name'    => static::CHECK_ENABLED,
//                        'is_bool' => true,
//                        'desc'    => $this->l('The verification by the web service is slow. Enabling this option can slow down your shop.'),
//                        'values'  => [
//                            [
//                                'id'    => 'active_on',
//                                'value' => 1,
//                                'label' => $this->l('Enabled'),
//                            ],
//                            [
//                                'id'    => 'active_off',
//                                'value' => 0,
//                                'label' => $this->l('Disabled'),
//                            ],
//                        ],
//                    ],
                ],
                'submit' => [
                    'title' => $this->l('Save'),
                ],
            ],
        ];

        $helper = new HelperForm();
        $helper->module = $this;
        $helper->show_toolbar = false;
        $helper->table = $this->table;
        $lang = new Language((int) Configuration::get('PS_LANG_DEFAULT'));
        $helper->default_form_language = $lang->id;
        $helper->allow_employee_form_lang = Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') ? Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG') : 0;
        $this->fields_form = [];

        $helper->identifier = $this->identifier;
        $helper->submit_action = 'submitVatNumber';
        $helper->currentIndex = $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name;
        $helper->token = Tools::getAdminTokenLite('AdminModules');
        $helper->tpl_vars = [
            'fields_value' => $this->getConfigFieldsValues(),
            'languages'    => $this->context->controller->getLanguages(),
            'id_language'  => $this->context->language->id,
        ];

        return $helper->generateForm([$fieldsForm]);
    }

    /**
     * @return array
     * @throws PrestaShopException
     */
    public function getConfigFieldsValues()
    {
        return [
            static::COUNTRY       => Tools::getValue(static::COUNTRY, Configuration::get(static::COUNTRY)),
            static::CHECK_ENABLED => Tools::getValue(static::CHECK_ENABLED, Configuration::get(static::CHECK_ENABLED)),
        ];
    }

    /**
     * @throws PrestaShopException
     */
    public function ajaxProcessSetIdGroup()
    {
        header('Content-Type: application/json; charset=utf-8');
        $iso = Tools::getValue('countryIso');
        $idGroup = (int) Tools::getValue('idGroup');
        if (!$iso) {
            die(json_encode([
                'success' => false,
                'message' => $this->l('Country ISO code is missing'),
            ]));
        }
        if (!Tools::isSubmit('idGroup')) {
            die(json_encode([
                'success' => false,
                'message' => $this->l('Group ID is missing'),
            ]));
        }

        $config = json_decode(Configuration::get(static::GROUPS_CONFIG), true);
        if (!is_array($config)) {
            $config = [];
        }
        $config[strtoupper($iso)] = (int) $idGroup;
        Configuration::updateValue(static::GROUPS_CONFIG, json_encode($config));

        die(json_encode([
            'success' => true,
        ]));
    }

    /**
     * Render the VAT groups list
     *
     * @return string HTML
     * @throws Exception
     * @throws PrestaShopDatabaseException
     * @throws PrestaShopException
     * @throws SmartyException
     * @since 1.0.0
     */
    protected function renderVatGroupsList()
    {
        Media::addJsDef([
            'vat_number_module_ajax' => $this->context->link->getAdminLink('AdminModules', true)."&configure={$this->name}&tab={$this->tab}",
        ]);
        $fieldsList = [
            'code'     => [
                'title'   => $this->l('Code'),
                'width'   => 'auto',
                'search'  => false,
                'orderby' => false,
            ],
            'name'     => [
                'title'   => $this->l('Name'),
                'width'   => 'auto',
                'search'  => false,
                'orderby' => false,
            ],
            'id_group' => [
                'title'           => $this->l('Target group after validation'),
                'width'           => 'auto',
                'search'          => false, 'orderby' => false,
                'remove_onclick'  => true,
                'callback_object' => 'VATNumber',
                'callback'        => 'displayGroup',
            ],
        ];
        $i = 1;
        $config = json_decode(Configuration::get(static::GROUPS_CONFIG), true);
        $results = array_map(function ($item) use (&$i, $config) {
            return array_merge($item, [
                'id'       => $i++,
                'id_group' => isset($config[strtoupper($item['code'])]) ? (int) $config[strtoupper($item['code'])] : 0,
            ]);
        }, static::$vatPrefixes);

        $helperList = new HelperList();
        $helperList->id = 1;
        $helperList->shopLinkType = false;
        $helperList->list_id = 'vatgroups';
        $helperList->identifier = 'id';
        $helperList->module = $this;
        $helperList->currentPage = 1;
        $helperList->_pagination = [100];
        $helperList->_default_pagination = 100;
        $helperList->page = 1;
        $helperList->_defaultOrderBy = 'id';
        $helperList->listTotal = count($results);
        $helperList->bulk_actions = [];
        $helperList->actions = [];
        $helperList->title = $this->l('Group configuration');
        $helperList->token = Tools::getAdminTokenLite('AdminModules');
        $helperList->table = 'group';
        $helperList->currentIndex = AdminController::$currentIndex.'&'.http_build_query([
                'configure' => $this->name,
            ]);

        return $helperList->generateList($results, $fieldsList);
    }

    /**
     * @param int   $id
     * @param array $tr
     *
     * @return string
     * @throws Exception
     * @throws PrestaShopException
     * @throws SmartyException
     */
    public static function displayGroup($id, $tr)
    {
        $context = Context::getContext();
        $module = Module::getInstanceByName('vatnumber');
        $groups = Group::getGroups($context->language->id);

        $context->smarty->assign([
            'id'     => $id,
            'tr'     => $tr,
            'groups' => array_combine(array_merge([0], array_column($groups, 'id_group')), array_merge([$module->l('Do not change')], array_column($groups, 'name'))),
        ]);

        $reflection = new ReflectionClass($module);

        return $module->display($reflection->getFileName(), 'views/templates/admin/displayGroup.tpl');
    }
}
